Bootstrap: docker
From: python@sha256:1c26c25390307b64e8ff73e7edf34b4fbeac59d41da41c08da28dc316a721899

%files
    ./dummy_step.py /dummy_step.py

%post
    # Create directories
    mkdir -p /input_data
    mkdir -p /extra_implementation_specific_input_data
    mkdir -p /results
    mkdir -p /diagnostics

    # Install Python packages with specific versions
    pip install pandas==2.1.2 pyarrow pyyaml

%environment
    export LC_ALL=C
    export PYTHONPATH=/app:$PYTHONPATH

%runscript
    python /dummy_step.py "$@"

%labels
    Author Patrick Nast
    Version v1.0
    Description Python Pandas Implementation

%startscript
    # These directories should be bound when running the container
    mkdir -p /results
    mkdir -p /diagnostics
    mkdir -p /input_data